# coding=utf8

from datetime import timedelta, datetime as dt
import time
from util import print_traceback, dumpjs, get_int_from_server_config, get_string_from_server_config, get_object_from_server_config, get_next_run, update_baseconfig_value, print_response
from cloudsight import CloudsightWorker
from cloudsight_config import parse_cloudsight_from_baseconfig
from easydb_api import easydb_authenticate, confirm_pending_tasks
from api import api_start_now


def easydb_server_start(easydb_context):
    easydb_context.register_callback(
        'process', {
            'name': 'easydb-auto-keyworder-plugin'
        })

    # endpoint: <server>/api/plugin/base/easydb-auto-keyworder-plugin/start_now
    easydb_context.register_callback(
        'api', {
            'name': 'start_now',
            'callback': 'start_now'
        })


def start_now(easydb_context, parameters):
    logger = easydb_context.get_logger('pf.plugin.base.auto_keyworder.api')
    logger.debug('[/start_now] parameters: %s' %
                 dumpjs(parameters))
    return api_start_now(easydb_context, logger, parameters)


def run(easydb_context):
    logger = easydb_context.get_logger('pf.plugin.base.auto_keyworder')

    # get server config
    server_config = None
    try:
        conn = easydb_context.db_connect('easydb-auto-keyworder-plugin')
        server_config = easydb_context.get_config(conn)
        conn.close()
    except Exception as e:
        logger.error('could not get server config: %s' % e.message)
        return

    search_chunk_size = get_int_from_server_config(
        server_config, 'system.auto_keyworder.search_chunk_size', 100)
    if search_chunk_size < 0:
        search_chunk_size = 0
    logger.debug(
        'server config: system.auto_keyworder.search_chunk_size=%d' % search_chunk_size)

    poll_start_hour = get_int_from_server_config(
        server_config, 'system.auto_keyworder.poll_start_hour', 0)
    if poll_start_hour < 0 or poll_start_hour > 23:
        poll_start_hour = 0
    logger.debug(
        'server config: system.auto_keyworder.poll_start_hour=%d' % poll_start_hour)

    poll_every_days = get_int_from_server_config(
        server_config, 'system.auto_keyworder.poll_every_days', 0)
    if poll_every_days < 0:
        poll_every_days = 0
    logger.debug(
        'server config: system.auto_keyworder.poll_every_days=%d' % poll_every_days)

    first_run = True
    next_run = calculate_next_runtime(poll_start_hour,
                                      poll_every_days,
                                      first_run,
                                      logger)
    logger.debug('[first run: %s] next run time for cloudsight worker: %s' % (
        first_run, next_run))

    while True:
        cloudsight_config = None
        conn = None
        try:
            # load server config, get base config, parse auto keyworder configurations from base config
            conn = easydb_context.db_connect('easydb-auto-keyworder-plugin')
            server_config = easydb_context.get_config(conn)
            conn.close()
            conn = None

            base_config = get_object_from_server_config(
                server_config, 'base.system.auto_keyworder')

            cloudsight_config = parse_cloudsight_from_baseconfig(
                easydb_context, base_config)

            if cloudsight_config is not None:
                if cloudsight_config.start_now:
                    logger.debug('start_now flag is True')
                    conn = easydb_context.db_connect('easydb-auto-keyworder-plugin')
                    cursor = conn.cursor()
                    if update_baseconfig_value(cursor,
                                               logger,
                                               _class='system',
                                               _key='auto_keyworder',
                                               _parameter='start_now',
                                               value_type='value_bool',
                                               value=False):
                        logger.debug(
                            'set baseconfig entry system.auto_keyworder.start_now to false')
                    else:
                        logger.debug(
                            'could not set baseconfig entry system.auto_keyworder.start_now to false')
                    conn.commit()

                    conn.close()
                    conn = None

                if cloudsight_config.start_now or next_run <= dt.utcnow():
                    run_cloudsight_worker(easydb_context,
                                          logger,
                                          server_config,
                                          cloudsight_config,
                                          search_chunk_size)
                    first_run = False
                    next_run = calculate_next_runtime(poll_start_hour,
                                                      poll_every_days,
                                                      first_run,
                                                      logger)
                    logger.debug('[first run: %s] next run time for cloudsight worker: %s' % (
                        first_run, next_run))

            else:
                logger.debug(
                    'no cloudsight config. will not start another update process')

        except Exception as e:
            if conn:
                conn.close()
            print_traceback(logger, e)

        baseconfig_poll_interval_sec = get_int_from_server_config(
            server_config, 'system.auto_keyworder.baseconfig_poll_interval_sec', 10)
        if baseconfig_poll_interval_sec < 1:
            baseconfig_poll_interval_sec = 10

        logger.debug('sleep for %d seconds' % baseconfig_poll_interval_sec)
        time.sleep(baseconfig_poll_interval_sec)


def run_cloudsight_worker(easydb_context, logger, server_config, cloudsight_config, search_chunk_size):

    logger.debug('setup cloudsight worker main loop')

    server = get_string_from_server_config(
        server_config, 'system.server.internal_url')
    if server is None:
        logger.warn(
            'could not get system.server.internal_url. try fallback system.server.external_url')
        server = get_string_from_server_config(
            server_config, 'system.server.external_url')
        if server is None:
            logger.error(
                'could not get system.server.external_url. stop')
            return False

    if not cloudsight_config.enabled:
        logger.info('cloudsight disabled')
        return False

    if len(cloudsight_config.cloudsight_config_list) < 1:
        logger.info('no enabled cloudsight configurations')
        return False

    # if login and password are set, authenticate at easydb to get a token to perform searches and update objects
    if cloudsight_config.easydb_session_token is None:
        if cloudsight_config.easydb_user is None:
            logger.error(
                'can not authenticate at easydb %s: easydb_user not set' % server)
            return False
        if cloudsight_config.easydb_password is None:
            logger.error(
                'can not authenticate at easydb %s: easydb_password not set' % server)
            return False

        logger.debug('authenticate at easydb %s: use login %s' %
                     (server, cloudsight_config.easydb_user))
        token, resp = easydb_authenticate(
            server,
            cloudsight_config.easydb_user,
            cloudsight_config.easydb_password
        )
        if resp is not None:
            logger.debug('Response of authenticate at easydb %s: %s' %
                         (server, print_response(resp)))

        if token is None:
            logger.error('Could not authenticate at easydb %s: %s' %
                         (server, print_response(resp)))
            return False
        else:
            if resp is not None:
                resp, err = confirm_pending_tasks(server, token, resp)
                if err is not None:
                    logger.error('Could not confirm pending tasks at easydb %s: %s' % (
                        server, print_response(err)))
                    return False
                elif resp is not None:
                    logger.debug('Confirmed pending tasks at easydb %s: %s' % (
                        server, print_response(resp)))

    else:
        token = cloudsight_config.easydb_session_token
        logger.debug(
            'do not authenticate with login, use session token from baseconfig')

    logger.debug('session token: %s' % token)

    cloudsight_worker = CloudsightWorker(easydb_context,
                                         server,
                                         token,
                                         cloudsight_config,
                                         search_chunk_size)

    logger.info('start cloudsight worker')
    cloudsight_worker.main_loop()

    return True


def calculate_next_runtime(poll_start_hour, poll_every_days, first_run, logger):
    # calculate the next run time for the cloudsight worker
    if first_run:
        # for the first run, calculate the next possible full hour
        return get_next_run(poll_start_hour)
    else:
        # calculate the next possible full hour after the specified number of poll_every_days pause
        return get_next_run(poll_start_hour, poll_every_days)
