#!/usr/bin/env python3

# This script generates produce_config_default.yml from information embedded in this
# file (see below). When changing the script, run the script:
#
# …/resources/baseconfig/fas$ ./gen_produce_conf.py
#
# and commit both gen_produce_conf.py and produce_config_default.yml.

import yaml

data = {
    'uploadallowunknown': True,
    'classes': {},
}

languages = ('de-DE', 'en-US')
l10n = {
    '*': {
        'small': ('250px', '250px'),
        'preview': ('1000px', '1000px'),
        'preview_watermark': ('1000px (Wasserzeichen)', '1000px (watermark)'),
        'huge': ('2000px', '2000px'),
        'full': ('Original (formatiert)', 'Original (formatted)'),
        'svg': ('SVG', 'SVG'),
    },
    "audio": {
        "aac": ("AAC", "AAC"),
    },
    'video': {
        'preview': ('720px', '720px'),
        'huge': ('1080px', '1080px'),
        '360p': ('360p', '360p'),
        '720p': ('720p', '720p'),
        '1080p': ('1080p', '1080p'),
    },
    'office': {
        'pages': ('(Seiten)', '(Pages)'),
    },
}

extensions = {
    'image': ['ai','arw','bmp','cr2','dng','eps','gif','heic','ico','jp2','jpeg','jpg','nef','png','ppm','psd',
        'raf','svg','tif','tiff','webp', 'wmf'],
    'office': ['csv','doc','docx','odp','ods','odt','pdf','ppt','pptx','rtf','txt','xls','xlsx'],
    'video': ['mp4','avi','mov','ogv','wmv','webm','3gp','m4v','asf','hevc','m2v','mjpeg','mkv','mxf','rm',
        'swf','ts','wtv','mts','mpeg','mpg','vob','flv'],
    'audio': ['aac','aiff','dts','flac','m4a','mp2','mp3','oga','ogg','opus','ra','wav','wma'],
}

def _reformat(fmt, size_minimum = None):
    d = {
        'recipename': 'imageconverter:reformat',
        'params': { 'format': fmt }}
    if size_minimum is not None:
        d['params']['size_minimum'] = size_minimum
    return d

def _browserthumbs(fmt, size, size_min = False, size_minimum = None, strip = None):
    d = {
        'recipename': 'imageconverter:browserthumbs',
        'params': { 'size': size, 'format': fmt }}
    if size_min:
        d['params']['resize_mode'] = 'min'
    if size_minimum is not None:
        d['params']['size_minimum'] = size_minimum
    if strip is not None:
        d['params']['strip'] = strip
    return d

def _poolwatermark(fmt, size):
    return {
        'recipename': 'imageconverter:preview_pool_watermark',
        'params': {
            'size': size,
            'format': fmt,
        }
    }

def _svg():
    return {
        'recipename': 'imageconverter:vectortosvg',
    }

def _officethumbs(fmt, size):
    d = {
        'recipename': 'officeconverter:browserthumbs',
        'params': { 'size': size, 'format': fmt }}
    return d

def _officepreview(size):
    return {
        'recipename': 'officeconverter:preview',
        'params': {
            'size': size,
            'format': 'jpg',
            'quality': 80
        }
    }

def _vresize(height,  height_minimum = None):
    d = {
        'recipename': 'videoconverter:resize',
        'params': {'height': height }}
    if height_minimum is not None:
        d['params']['height_minimum'] = height_minimum
    return d

def _vpreview(fmt, size):
    return {
        'recipename': 'videoconverter:thumbnail',
        'params': {
            'size': '%s' % size,
            'format': fmt,
        }
    }

def _apreview(fmt, size):
    return {
        'recipename': 'audioconverter:preview',
        'params': {
            'size': size,
            'format': fmt,
        }
    }

def _aconvert(fmt):
    return {
        'recipename': 'audioconverter:convert',
        'params': {
            'format': fmt,
        }
    }

versions = {
    'image': {
        'jpg,jpeg': {
            'small': _browserthumbs('jpg', 250, True, None, True),
            'preview': _browserthumbs('jpg', 1000, False, 251, True),
            'preview_watermark': _poolwatermark('jpg', 1000),
            'huge': _browserthumbs('jpg', 2000, False, 1001, True),
            'full': _reformat('jpg', 2001),
        },
        'wmf': {
            'small': _browserthumbs('png', 250, True, None, True),
            'preview': _browserthumbs('png', 1000, False, 251, True),
            'preview_watermark': _poolwatermark('png', 1000),
            'huge': _browserthumbs('png', 2000, False, 1001, True),
            'svg': _svg(),
        },
        '*': {
            'small': _browserthumbs('png', 250, True, None, True),
            'preview': _browserthumbs('png', 1000, False, 251, True),
            'preview_watermark': _poolwatermark('png', 1000),
            'huge': _browserthumbs('png', 2000, False, 1001, True),
            'full': _reformat('png', 2001),
        },
    },
    'office': {
        'pdf': {
            'small': _officethumbs('jpg', 250),
            'preview': _officethumbs('jpg', 1000),
            'preview_watermark': _poolwatermark('jpg', 1000),
            'pages': { 'recipename': 'pdfconverter:pdfpages' },
        },
        '*': {
            'small': _officepreview(250),
            'preview': _officepreview(1000),
            'preview_watermark': _poolwatermark('png', 1000),
        }
    },
    'video': {
        '*': {
            'small': _vpreview('jpg', 250),
            'preview': _vpreview('jpg', 720),
            'preview_watermark': _poolwatermark('jpg', 720),
            'huge': _vpreview('jpg', 1080),
            '360p': _vresize(360),
            '720p': _vresize(720, 361),
            '1080p': _vresize(1080, 721),
        }
    },
    'audio': {
        # preview images, directly playable in browser
        'flac,m4a,mp3,oga,ogg,opus': {
            'small': _apreview('jpg', 250),
            'preview': _apreview('jpg', 1000),
            'preview_watermark': _poolwatermark('jpg', 1000),
        },
        # preview images & aac version to play in browser
        "aiff,mp2,wma": {
            'small': _apreview('jpg', 250),
            'preview': _apreview('jpg', 1000),
            'preview_watermark': _poolwatermark('jpg', 1000),
            'aac': _aconvert('aac'),
        },
        # only aac version to play in browser
        'dts,ra,wav': {
            'aac': _aconvert('aac'),
        },
        '*': {
            #'aac':
        },
    },
}

groups = {
    'small': 'thumbnail',
    'preview': 'preview',
    'preview_watermark': 'preview',
    'huge': 'huge',
    'full': 'huge',
    '360p': 'preview',
    '720p': 'preview',
    '1080p': 'huge',
    'aac': None,
}

def get_source_version(versionname, srcexts):
    sourceversions = {
        'preview_watermark': 'preview',
    }
    if versionname in sourceversions:
        return sourceversions[versionname]
    if srcexts == 'wmf' and versionname != 'svg':
        return 'svg'
    return ''

versions_standard = [ 'small', 'preview', ]
versions_watermark = [ 'preview_watermark', ]
versions_no_rm = [ 'small', ]
for cls, extspecs in versions.items():
    cdata = {
        'uploadenabledextensions': extensions[cls],
        'versions': [],
    }

    specialized_exts = set()
    for extspec, verspecs in extspecs.items():
        for vername, verspec in verspecs.items():
            vdata = {
                'name': vername,
                'displayname': {},
                'group': groups.get(vername),
                'sourceversion': get_source_version(vername, extspec),
                'rightsmanagement': vername not in versions_no_rm,
                'standard': vername in versions_standard,
                'watermark': vername in versions_watermark,
                'extensions': extspec == '*' and
                    [ x for x in extensions[cls] if x not in specialized_exts ] or
                    extspec.split(','),
                'recipename': verspec['recipename'],
                'params': verspec.get('params'),
            }
            l = l10n.get(cls, {}).get(vername)
            if l is None:
                l = l10n['*'][vername]
            for i in range(len(languages)):
                vdata['displayname'][languages[i]] = l[i]
            cdata['versions'].append(vdata)
        if extspec != '*':
            specialized_exts |= set(extspec.split(','))

    data['classes'][cls] = cdata

with open('produce_config_default.yml', 'w') as f:
    f.write('# file is generated by gen_produce_conf.py, do not modify manually\n')
    f.write(yaml.dump(data, default_flow_style = False))
